@@ -60,14 +60,28 @@ end
 # #private_key        :: The client's SSL private key
 # #reuse_ssl_sessions :: Reuse a previously opened SSL session for a new
 #                        connection
+# #ssl_version        :: Which specific SSL version to use
 # #verify_callback    :: For server certificate verification
 # #verify_mode        :: How connections should be verified
 #
 # == Proxies
 #
-# A proxy can be used either by providing a URI as the second argument of
-# ::new or by giving <code>:ENV</code> as the second argument which will
-# consult environment variables.  See proxy_from_env for details.
+# A proxy can be set through #proxy= or at initialization time by providing a
+# second argument to ::new.  The proxy may be the URI of the proxy server or
+# <code>:ENV</code> which will consult environment variables.
+#
+# See #proxy= and #proxy_from_env for details.
+#
+# == Headers
+#
+# Headers may be specified for use in every request.  #headers are appended to
+# any headers on the request.  #override_headers replace existing headers on
+# the request.
+#
+# The difference between the two can be seen in setting the User-Agent.  Using
+# <code>http.headers['User-Agent'] = 'MyUserAgent'</code> will send "Ruby,
+# MyUserAgent" while <code>http.override_headers['User-Agent'] =
+# 'MyUserAgent'</code> will send "MyUserAgent".
 #
 # == Tuning
 #
@@ -96,6 +110,14 @@ end
 # The amount of time to wait for a connection to be opened.  Set through
 # #open_timeout.
 #
+# === Idle Timeout
+#
+# If a connection has not been used in this many seconds it will be reset when
+# a request would use the connection.  The default idle timeout is unlimited.
+# If you know the server's idle timeout setting this value will eliminate
+# failures from attempting non-idempotent requests on closed connections.  Set
+# through #idle_timeout.
+#
 # === Socket Options
 #
 # Socket options may be set on newly-created connections.  See #socket_options
@@ -107,8 +129,9 @@ end
 # setting retry_change_requests to true requests will automatically be retried
 # once.
 #
-# Only do this when you know that retrying a POST is safe for your
-# application and will not create duplicate resources.
+# Only do this when you know that retrying a POST or other non-idempotent
+# request is safe for your application and will not create duplicate
+# resources.
 #
 # The recommended way to handle non-idempotent requests is the following:
 #
@@ -149,7 +172,7 @@ class Net::HTTP::Persistent
   ##
   # The version of Net::HTTP::Persistent you are using
 
-  VERSION = '2.3.3'
+  VERSION = '2.5.2'
 
   ##
   # Error class for errors raised by Net::HTTP::Persistent.  Various
@@ -161,24 +184,19 @@ class Net::HTTP::Persistent
   ##
   # This client's OpenSSL::X509::Certificate
 
-  attr_accessor :certificate
+  attr_reader :certificate
 
   ##
   # An SSL certificate authority.  Setting this will set verify_mode to
   # VERIFY_PEER.
 
-  attr_accessor :ca_file
+  attr_reader :ca_file
 
   ##
-  # An SSL certificate store. Setting this will override the default
-  # certificate store. See verify_mode for more information.
+  # An SSL certificate store.  Setting this will override the default
+  # certificate store.  See verify_mode for more information.
 
-  attr_accessor :cert_store
-
-  ##
-  # Where this instance's connections live in the thread local variables
-
-  attr_reader :connection_key # :nodoc:
+  attr_reader :cert_store
 
   ##
   # Sends debug_output to this IO via Net::HTTP#set_debug_output.
@@ -189,7 +207,17 @@ class Net::HTTP::Persistent
   attr_accessor :debug_output
 
   ##
-  # Headers that are added to every request
+  # Current connection generation
+
+  attr_reader :generation # :nodoc:
+
+  ##
+  # Where this instance's connections live in the thread local variables
+
+  attr_reader :generation_key # :nodoc:
+
+  ##
+  # Headers that are added to every request using Net::HTTP#add_field
 
   attr_reader :headers
 
@@ -227,10 +255,15 @@ class Net::HTTP::Persistent
 
   attr_accessor :open_timeout
 
+  ##
+  # Headers that are added to every request using Net::HTTP#[]=
+
+  attr_reader :override_headers
+
   ##
   # This client's SSL private key
 
-  attr_accessor :private_key
+  attr_reader :private_key
 
   ##
   # The URL through which requests will be proxied
@@ -267,6 +300,24 @@ class Net::HTTP::Persistent
 
   attr_reader :socket_options
 
+  ##
+  # Current SSL connection generation
+
+  attr_reader :ssl_generation # :nodoc:
+
+  ##
+  # Where this instance's SSL connections live in the thread local variables
+
+  attr_reader :ssl_generation_key # :nodoc:
+
+  ##
+  # SSL version to use.
+  #
+  # By default, the version will be negotiated automatically between client
+  # and server.  Ruby 1.9 and newer only.
+
+  attr_reader :ssl_version if RUBY_VERSION > '1.9'
+
   ##
   # Where this instance's last-use times live in the thread local variables
 
@@ -275,21 +326,18 @@ class Net::HTTP::Persistent
   ##
   # SSL verification callback.  Used when ca_file is set.
 
-  attr_accessor :verify_callback
+  attr_reader :verify_callback
 
   ##
   # HTTPS verify mode.  Defaults to OpenSSL::SSL::VERIFY_PEER which verifies
   # the server certificate.
   #
-  # If no certificate, ca_file or cert_store is set the default system
-  # certificate store is used.
-  #
-  # To disable server certificate validation set to OpenSSL::SSL::VERIFY_NONE,
-  # but this is a bad idea as it disables SSL protections.
+  # If no ca_file or cert_store is set the default system certificate store is
+  # used.
   #
   # You can use +verify_mode+ to override any default values.
 
-  attr_accessor :verify_mode
+  attr_reader :verify_mode
 
   ##
   # Enable retries of non-idempotent requests that change data (e.g. POST
@@ -322,61 +370,116 @@ class Net::HTTP::Persistent
   def initialize name = nil, proxy = nil
     @name = name
 
-    @proxy_uri = case proxy
-                 when :ENV      then proxy_from_env
-                 when URI::HTTP then proxy
-                 when nil       then # ignore
-                 else raise ArgumentError, 'proxy must be :ENV or a URI::HTTP'
-                 end
-
-    if @proxy_uri then
-      @proxy_args = [
-        @proxy_uri.host,
-        @proxy_uri.port,
-        @proxy_uri.user,
-        @proxy_uri.password,
-      ]
-
-      @proxy_connection_id = [nil, *@proxy_args].join ':'
-    end
-
-    @debug_output   = nil
-    @headers        = {}
-    @http_versions  = {}
-    @keep_alive     = 30
-    @open_timeout   = nil
-    @read_timeout   = nil
-    @idle_timeout   = 5
-    @socket_options = []
+    @debug_output     = nil
+    @proxy_uri        = nil
+    @headers          = {}
+    @override_headers = {}
+    @http_versions    = {}
+    @keep_alive       = 30
+    @open_timeout     = nil
+    @read_timeout     = nil
+    @idle_timeout     = 5
+    @socket_options   = []
 
     @socket_options << [Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1] if
       Socket.const_defined? :TCP_NODELAY
 
     key = ['net_http_persistent', name].compact
-    @connection_key = [key, 'connections'].join('_').intern
-    @request_key    = [key, 'requests'].join('_').intern
-    @timeout_key    = [key, 'timeouts'].join('_').intern
+    @generation_key     = [key, 'generations'    ].join('_').intern
+    @ssl_generation_key = [key, 'ssl_generations'].join('_').intern
+    @request_key        = [key, 'requests'       ].join('_').intern
+    @timeout_key        = [key, 'timeouts'       ].join('_').intern
 
     @certificate        = nil
     @ca_file            = nil
     @private_key        = nil
+    @ssl_version        = nil
     @verify_callback    = nil
     @verify_mode        = OpenSSL::SSL::VERIFY_PEER
     @cert_store         = nil
+
+    @generation         = 0 # incremented when proxy URI changes
+    @ssl_generation     = 0 # incremented when SSL session variables change
     @reuse_ssl_sessions = true
 
     @retry_change_requests = false
+
+    self.proxy = proxy if proxy
+  end
+
+  ##
+  # Sets this client's OpenSSL::X509::Certificate
+
+  def certificate= certificate
+    @certificate = certificate
+
+    reconnect_ssl
+  end
+
+  ##
+  # Sets the SSL certificate authority file.
+
+  def ca_file= file
+    @ca_file = file
+
+    reconnect_ssl
+  end
+
+  ##
+  # Overrides the default SSL certificate store used for verifying
+  # connections.
+
+  def cert_store= store
+    @cert_store = store
+
+    reconnect_ssl
+  end
+
+  ##
+  # Finishes all connections on the given +thread+ that were created before
+  # the given +generation+ in the threads +generation_key+ list.
+  #
+  # See #shutdown for a bunch of scary warning about misusing this method.
+
+  def cleanup(generation, thread = Thread.current,
+              generation_key = @generation_key) # :nodoc:
+    timeouts = thread[@timeout_key]
+
+    (0...generation).each do |old_generation|
+      conns = thread[generation_key].delete old_generation
+
+      conns.each_value do |conn|
+        finish conn, thread
+
+        timeouts.delete conn.object_id if timeouts
+      end if conns
+    end
   end
 
   ##
   # Creates a new connection for +uri+
 
   def connection_for uri
-    Thread.current[@connection_key] ||= {}
-    Thread.current[@request_key]    ||= Hash.new 0
-    Thread.current[@timeout_key]    ||= Hash.new EPOCH
+    Thread.current[@generation_key]     ||= Hash.new { |h,k| h[k] = {} }
+    Thread.current[@ssl_generation_key] ||= Hash.new { |h,k| h[k] = {} }
+    Thread.current[@request_key]        ||= Hash.new 0
+    Thread.current[@timeout_key]        ||= Hash.new EPOCH
 
-    connections = Thread.current[@connection_key]
+    use_ssl = uri.scheme.downcase == 'https'
+
+    if use_ssl then
+      ssl_generation = @ssl_generation
+
+      ssl_cleanup ssl_generation
+
+      connections = Thread.current[@ssl_generation_key][ssl_generation]
+    else
+      generation = @generation
+
+      cleanup generation
+
+      connections = Thread.current[@generation_key][generation]
+    end
 
     net_http_args = [uri.host, uri.port]
     connection_id = net_http_args.join ':'
@@ -391,7 +494,7 @@ class Net::HTTP::Persistent
     unless connection = connections[connection_id] then
       connections[connection_id] = http_class.new(*net_http_args)
       connection = connections[connection_id]
-      ssl connection if uri.scheme.downcase == 'https'
+      ssl connection if use_ssl
     else
       last_used = Thread.current[@timeout_key][connection.object_id]
       reset connection unless last_used > max_age
@@ -444,15 +547,22 @@ class Net::HTTP::Persistent
   ##
   # Finishes the Net::HTTP +connection+
 
-  def finish connection
-    Thread.current[@request_key].delete connection.object_id
+  def finish connection, thread = Thread.current
+    if requests = thread[@request_key] then
+      requests.delete connection.object_id
+    end
 
     connection.finish
   rescue IOError
   end
 
   def http_class # :nodoc:
-    @reuse_ssl_sessions ? Net::HTTP::Persistent::SSLReuse : Net::HTTP
+    if [:FakeWeb, :WebMock].any? { |klass| Object.const_defined?(klass) } or
+      not @reuse_ssl_sessions then
+        Net::HTTP
+    else
+      Net::HTTP::Persistent::SSLReuse
+    end
   end
 
   ##
@@ -543,6 +653,50 @@ class Net::HTTP::Persistent
     connection.pipeline requests, &block
   end
 
+  ##
+  # Sets this client's SSL private key
+
+  def private_key= key
+    @private_key = key
+
+    reconnect_ssl
+  end
+
+  ##
+  # Sets the proxy server.  The +proxy+ may be the URI of the proxy server,
+  # the symbol +:ENV+ which will read the proxy from the environment or nil to
+  # disable use of a proxy.  See #proxy_from_env for details on setting the
+  # proxy from the environment.
+  #
+  # If the proxy URI is set after requests have been made, the next request
+  # will shut-down and re-open all connections.
+  #
+  # If you are making some requests through a proxy and others without a proxy
+  # use separate Net::Http::Persistent instances.
+
+  def proxy= proxy
+    @proxy_uri = case proxy
+                 when :ENV      then proxy_from_env
+                 when URI::HTTP then proxy
+                 when nil       then # ignore
+                 else raise ArgumentError, 'proxy must be :ENV or a URI::HTTP'
+                 end
+
+    if @proxy_uri then
+      @proxy_args = [
+        @proxy_uri.host,
+        @proxy_uri.port,
+        @proxy_uri.user,
+        @proxy_uri.password,
+      ]
+
+      @proxy_connection_id = [nil, *@proxy_args].join ':'
+    end
+
+    reconnect
+    reconnect_ssl
+  end
+
   ##
   # Creates a URI for an HTTP proxy server from ENV variables.
   #
@@ -552,7 +706,7 @@ class Net::HTTP::Persistent
   # indicated user and password unless HTTP_PROXY contains either of these in
   # the URI.
   #
-  # For Windows users lowercase ENV variables are preferred over uppercase ENV
+  # For Windows users, lowercase ENV variables are preferred over uppercase ENV
   # variables.
 
   def proxy_from_env
@@ -570,6 +724,20 @@ class Net::HTTP::Persistent
     uri
   end
 
+  ##
+  # Forces reconnection of HTTP connections.
+
+  def reconnect
+    @generation += 1
+  end
+
+  ##
+  # Forces reconnection of SSL connections.
+
+  def reconnect_ssl
+    @ssl_generation += 1
+  end
+
   ##
   # Finishes then restarts the Net::HTTP +connection+
 
@@ -604,10 +772,14 @@ class Net::HTTP::Persistent
 
     req = Net::HTTP::Get.new uri.request_uri unless req
 
-    headers.each do |pair|
+    @headers.each do |pair|
       req.add_field(*pair)
     end
 
+    @override_headers.each do |name, value|
+      req[name] = value
+    end
+
     unless req['Connection'] then
       req.add_field 'Connection', 'keep-alive'
       req.add_field 'Keep-Alive', @keep_alive
@@ -676,17 +848,14 @@ class Net::HTTP::Persistent
   # best to call #shutdown in the thread at the appropriate time instead!
 
   def shutdown thread = Thread.current
-    connections = thread[@connection_key]
+    generation = reconnect
+    cleanup generation, thread, @generation_key
 
-    connections.each do |_, connection|
-      begin
-        connection.finish
-      rescue IOError
-      end
-    end if connections
+    ssl_generation = reconnect_ssl
+    cleanup ssl_generation, thread, @ssl_generation_key
 
-    thread[@connection_key] = nil
-    thread[@request_key]    = nil
+    thread[@request_key] = nil
+    thread[@timeout_key] = nil
   end
 
   ##
@@ -713,8 +882,37 @@ class Net::HTTP::Persistent
   def ssl connection
     connection.use_ssl = true
 
+    connection.ssl_version = @ssl_version if @ssl_version
+
     connection.verify_mode = @verify_mode
 
+    if OpenSSL::SSL::VERIFY_PEER == OpenSSL::SSL::VERIFY_NONE and
+       not Object.const_defined?(:I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG) then
+      warn <<-WARNING
+                             !!!SECURITY WARNING!!!
+
+The SSL HTTP connection to:
+
+  #{connection.address}:#{connection.port}
+
+                           !!!MAY NOT BE VERIFIED!!!
+
+On your platform your OpenSSL implementation is broken.
+
+There is no difference between the values of VERIFY_NONE and VERIFY_PEER.
+
+This means that attempting to verify the security of SSL connections may not
+work.  This exposes you to man-in-the-middle exploits, snooping on the
+contents of your connection and other dangers to the security of your data.
+
+To disable this warning define the following constant at top-level in your
+application:
+
+  I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG = nil
+
+      WARNING
+    end
+
     if @ca_file then
       connection.ca_file = @ca_file
       connection.verify_mode = OpenSSL::SSL::VERIFY_PEER
@@ -735,6 +933,45 @@ class Net::HTTP::Persistent
                             end
   end
 
+  ##
+  # Finishes all connections that existed before the given SSL parameter
+  # +generation+.
+
+  def ssl_cleanup generation # :nodoc:
+    cleanup generation, Thread.current, @ssl_generation_key
+  end
+
+  ##
+  # SSL version to use
+
+  def ssl_version= ssl_version
+    @ssl_version = ssl_version
+
+    reconnect_ssl
+  end if RUBY_VERSION > '1.9'
+
+  ##
+  # Sets the HTTPS verify mode.  Defaults to OpenSSL::SSL::VERIFY_PEER.
+  #
+  # Setting this to VERIFY_NONE is a VERY BAD IDEA and should NEVER be used.
+  # Securely transfer the correct certificate and update the default
+  # certificate store or set the ca file instead.
+
+  def verify_mode= verify_mode
+    @verify_mode = verify_mode
+
+    reconnect_ssl
+  end
+
+  ##
+  # SSL verification callback.
+
+  def verify_callback= callback
+    @verify_callback = callback
+
+    reconnect_ssl
+  end
+
 end
 
 require 'net/http/persistent/ssl_reuse'
