@@ -92,8 +92,10 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
   end
 
   def basic_connection
+    raise "#{@uri} is not HTTP" unless @uri.scheme.downcase == 'http'
+
     c = BasicConnection.new
-    conns["#{@uri.host}:#{@uri.port}"] = c
+    conns[0]["#{@uri.host}:#{@uri.port}"] = c
     c
   end
 
@@ -115,13 +117,24 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
   end
 
   def conns
-    Thread.current[@http.connection_key] ||= {}
+    Thread.current[@http.generation_key] ||= Hash.new { |h,k| h[k] = {} }
   end
 
   def reqs
     Thread.current[@http.request_key] ||= {}
   end
 
+  def ssl_conns
+    Thread.current[@http.ssl_generation_key] ||= Hash.new { |h,k| h[k] = {} }
+  end
+
+  def ssl_connection generation = 0
+    raise "#{@uri} is not HTTPS" unless @uri.scheme.downcase == 'https'
+    c = BasicConnection.new
+    ssl_conns[generation]["#{@uri.host}:#{@uri.port}"] = c
+    c
+  end
+
   def touts
     Thread.current[@http.timeout_key] ||= Hash.new Net::HTTP::Persistent::EPOCH
   end
@@ -135,14 +148,7 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_equal 'name', http.name
   end
 
-  def test_initialize_env
-    ENV['HTTP_PROXY'] = 'proxy.example'
-    http = Net::HTTP::Persistent.new nil, :ENV
-
-    assert_equal URI.parse('http://proxy.example'), http.proxy_uri
-  end
-
-  def test_initialize_uri
+  def test_initialize_proxy
     proxy_uri = URI.parse 'http://proxy.example'
 
     http = Net::HTTP::Persistent.new nil, proxy_uri
@@ -150,6 +156,27 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_equal proxy_uri, http.proxy_uri
   end
 
+  def test_ca_file_equals
+    @http.ca_file = :ca_file
+
+    assert_equal :ca_file, @http.ca_file
+    assert_equal 1, @http.ssl_generation
+  end
+
+  def test_cert_store_equals
+    @http.cert_store = :cert_store
+
+    assert_equal :cert_store, @http.cert_store
+    assert_equal 1, @http.ssl_generation
+  end
+
+  def test_certificate_equals
+    @http.certificate = :cert
+
+    assert_equal :cert, @http.certificate
+    assert_equal 1, @http.ssl_generation
+  end
+
   def test_connection_for
     @http.open_timeout = 123
     @http.read_timeout = 321
@@ -163,8 +190,8 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_equal 123, c.open_timeout
     assert_equal 321, c.read_timeout
 
-    assert_includes conns.keys, 'example.com:80'
-    assert_same c, conns['example.com:80']
+    assert_includes conns[0].keys, 'example.com:80'
+    assert_same c, conns[0]['example.com:80']
 
     socket = c.instance_variable_get :@socket
     expected = if Socket.const_defined? :TCP_NODELAY then
@@ -179,7 +206,7 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
   def test_connection_for_cached
     cached = basic_connection
     cached.start
-    conns['example.com:80'] = cached
+    conns[0]['example.com:80'] = cached
 
     c = @http.connection_for @uri
 
@@ -203,7 +230,7 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert c.started?
 
     assert_includes conns.keys, 'example.com:80'
-    assert_same c, conns['example.com:80']
+    assert_same c, conns[0]['example.com:80']
 
     socket = c.instance_variable_get :@socket
 
@@ -220,8 +247,8 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert c.started?
     assert_equal io, c.instance_variable_get(:@debug_output)
 
-    assert_includes conns.keys, 'example.com:80'
-    assert_same c, conns['example.com:80']
+    assert_includes conns[0].keys, 'example.com:80'
+    assert_same c, conns[0]['example.com:80']
   end
 
   def test_connection_for_finished_ssl
@@ -244,7 +271,7 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     cached = basic_connection
     def cached.start; raise Errno::EHOSTDOWN end
     def cached.started?; false end
-    conns['example.com:80'] = cached
+    conns[0]['example.com:80'] = cached
 
     e = assert_raises Net::HTTP::Persistent::Error do
       @http.connection_for @uri
@@ -253,6 +280,26 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_match %r%host down%, e.message
   end
 
+  def test_connection_for_http_class_with_fakeweb
+    Object.send :const_set, :FakeWeb, nil
+    c = @http.connection_for @uri
+    assert_instance_of Net::HTTP, c
+  ensure
+    if Object.const_defined?(:FakeWeb) then
+      Object.send :remove_const, :FakeWeb
+    end
+  end
+
+  def test_connection_for_http_class_with_webmock
+    Object.send :const_set, :WebMock, nil
+    c = @http.connection_for @uri
+    assert_instance_of Net::HTTP, c
+  ensure
+    if Object.const_defined?(:WebMock) then
+      Object.send :remove_const, :WebMock
+    end
+  end
+
   def test_connection_for_name
     http = Net::HTTP::Persistent.new 'name'
     uri = URI.parse 'http://example/'
@@ -285,16 +332,16 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert c.started?
     assert c.proxy?
 
-    assert_includes conns.keys,
+    assert_includes conns[1].keys,
                     'example.com:80:proxy.example:80:johndoe:muffins'
-    assert_same c, conns['example.com:80:proxy.example:80:johndoe:muffins']
+    assert_same c, conns[1]['example.com:80:proxy.example:80:johndoe:muffins']
   end
 
   def test_connection_for_refused
     cached = basic_connection
     def cached.start; raise Errno::ECONNREFUSED end
     def cached.started?; false end
-    conns['example.com:80'] = cached
+    conns[0]['example.com:80'] = cached
 
     e = assert_raises Net::HTTP::Persistent::Error do
       @http.connection_for @uri
@@ -325,6 +372,28 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert c.use_ssl?
   end
 
+  def test_connection_for_ssl_cached
+    @uri = URI.parse 'https://example.com/path'
+
+    cached = ssl_connection 0
+
+    c = @http.connection_for @uri
+
+    assert_same cached, c
+  end
+
+  def test_connection_for_ssl_cached_reconnect
+    @uri = URI.parse 'https://example.com/path'
+
+    cached = ssl_connection
+
+    @http.reconnect_ssl
+
+    c = @http.connection_for @uri
+
+    refute_same cached, c
+  end
+
   def test_connection_for_ssl_case
     uri = URI.parse 'HTTPS://example.com/path'
     c = @http.connection_for uri
@@ -338,7 +407,7 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     cached.start
     reqs[cached.object_id] = 10
     touts[cached.object_id] = Time.now - 6
-    conns['example.com:80'] = cached
+    conns[0]['example.com:80'] = cached
 
     c = @http.connection_for @uri
 
@@ -407,14 +476,24 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     refute @http.idempotent? Net::HTTP::Post.new '/'
   end
 
+  def test_max_age
+    assert_in_delta Time.now - 5, @http.max_age
+  end
+
   def test_normalize_uri
     assert_equal 'http://example',  @http.normalize_uri('example')
     assert_equal 'http://example',  @http.normalize_uri('http://example')
     assert_equal 'https://example', @http.normalize_uri('https://example')
   end
 
-  def test_max_age
-    assert_in_delta Time.now - 5, @http.max_age
+  def test_override_haeders
+    assert_empty @http.override_headers
+
+    @http.override_headers['User-Agent'] = 'MyCustomAgent'
+
+    expected = { 'User-Agent' => 'MyCustomAgent' }
+
+    assert_equal expected, @http.override_headers
   end
 
   def test_pipeline
@@ -436,6 +515,32 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_equal '/2', responses.last
   end
 
+  def test_private_key_equals
+    @http.private_key = :private_key
+
+    assert_equal :private_key, @http.private_key
+    assert_equal 1, @http.ssl_generation
+  end
+
+  def test_proxy_equals_env
+    ENV['HTTP_PROXY'] = 'proxy.example'
+
+    @http.proxy = :ENV
+
+    assert_equal URI.parse('http://proxy.example'), @http.proxy_uri
+
+    assert_equal 1, @http.generation, 'generation'
+    assert_equal 1, @http.ssl_generation, 'ssl_generation'
+  end
+
+  def test_proxy_equals_uri
+    proxy_uri = URI.parse 'http://proxy.example'
+
+    @http.proxy = proxy_uri
+
+    assert_equal proxy_uri, @http.proxy_uri
+  end
+
   def test_proxy_from_env
     ENV['HTTP_PROXY']      = 'proxy.example'
     ENV['HTTP_PROXY_USER'] = 'johndoe'
@@ -476,73 +581,21 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_nil uri
   end
 
-  def test_reset
-    c = basic_connection
-    c.start
-    touts[c.object_id] = Time.now
-    reqs[c.object_id]  = 5
+  def test_reconnect
+    result = @http.reconnect
 
-    @http.reset c
-
-    assert c.started?
-    assert c.finished?
-    assert c.reset?
-    assert_nil reqs[c.object_id]
-    assert_equal Net::HTTP::Persistent::EPOCH, touts[c.object_id]
+    assert_equal 1, result
   end
 
-  def test_reset_io_error
-    c = basic_connection
-    touts[c.object_id] = Time.now
-    reqs[c.object_id] = 5
+  def test_reconnect_ssl
+    result = @http.reconnect_ssl
 
-    @http.reset c
-
-    assert c.started?
-    assert c.finished?
-  end
-
-  def test_reset_host_down
-    c = basic_connection
-    touts[c.object_id] = Time.now
-    def c.start; raise Errno::EHOSTDOWN end
-    reqs[c.object_id] = 5
-
-    e = assert_raises Net::HTTP::Persistent::Error do
-      @http.reset c
-    end
-
-    assert_match %r%host down%, e.message
-  end
-
-  def test_reset_refused
-    c = basic_connection
-    touts[c.object_id] = Time.now
-    def c.start; raise Errno::ECONNREFUSED end
-    reqs[c.object_id] = 5
-
-    e = assert_raises Net::HTTP::Persistent::Error do
-      @http.reset c
-    end
-
-    assert_match %r%connection refused%, e.message
-  end
-
-  def test_ssl_error
-    uri = URI.parse 'https://example.com/path'
-    c = @http.connection_for uri
-    def c.request(*)
-      raise OpenSSL::SSL::SSLError, "SSL3_WRITE_PENDING:bad write retry"
-    end
-
-    e = assert_raises Net::HTTP::Persistent::Error do
-      @http.request uri
-    end
-    assert_match %r%bad write retry%, e.message
+    assert_equal 1, result
   end
 
   def test_request
-    @http.headers['user-agent'] = 'test ua'
+    @http.override_headers['user-agent'] = 'test ua'
+    @http.headers['accept'] = 'text/*'
     c = connection
 
     res = @http.request @uri
@@ -552,9 +605,12 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
 
     assert_kind_of Net::HTTP::Get, req
     assert_equal '/path',      req.path
+
+    assert_equal 'test ua',    req['user-agent']
+    assert_match %r%text/\*%,  req['accept']
+
     assert_equal 'keep-alive', req['connection']
     assert_equal '30',         req['keep-alive']
-    assert_match %r%test ua%,  req['user-agent']
 
     assert_in_delta Time.now, touts[c.object_id]
 
@@ -744,6 +800,17 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert c.finished?
   end
 
+  def test_request_post
+    c = connection
+
+    post = Net::HTTP::Post.new @uri.path
+
+    @http.request @uri, post
+    req = c.req
+
+    assert_same post, req
+  end
+
   def test_request_reset
     c = basic_connection
     def c.request(*a) raise Errno::ECONNRESET end
@@ -796,21 +863,94 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_match %r%too many connection resets%, e.message
   end
 
-  def test_request_post
-    c = connection
+  def test_request_ssl_error
+    uri = URI.parse 'https://example.com/path'
+    c = @http.connection_for uri
+    def c.request(*)
+      raise OpenSSL::SSL::SSLError, "SSL3_WRITE_PENDING:bad write retry"
+    end
 
-    post = Net::HTTP::Post.new @uri.path
+    e = assert_raises Net::HTTP::Persistent::Error do
+      @http.request uri
+    end
+    assert_match %r%bad write retry%, e.message
+  end
 
-    @http.request @uri, post
-    req = c.req
+  def test_reset
+    c = basic_connection
+    c.start
+    touts[c.object_id] = Time.now
+    reqs[c.object_id]  = 5
 
-    assert_same post, req
+    @http.reset c
+
+    assert c.started?
+    assert c.finished?
+    assert c.reset?
+    assert_nil reqs[c.object_id]
+    assert_equal Net::HTTP::Persistent::EPOCH, touts[c.object_id]
+  end
+
+  def test_reset_host_down
+    c = basic_connection
+    touts[c.object_id] = Time.now
+    def c.start; raise Errno::EHOSTDOWN end
+    reqs[c.object_id] = 5
+
+    e = assert_raises Net::HTTP::Persistent::Error do
+      @http.reset c
+    end
+
+    assert_match %r%host down%, e.message
+  end
+
+  def test_reset_io_error
+    c = basic_connection
+    touts[c.object_id] = Time.now
+    reqs[c.object_id] = 5
+
+    @http.reset c
+
+    assert c.started?
+    assert c.finished?
+  end
+
+  def test_reset_refused
+    c = basic_connection
+    touts[c.object_id] = Time.now
+    def c.start; raise Errno::ECONNREFUSED end
+    reqs[c.object_id] = 5
+
+    e = assert_raises Net::HTTP::Persistent::Error do
+      @http.reset c
+    end
+
+    assert_match %r%connection refused%, e.message
+  end
+
+  def test_retry_change_requests_equals
+    get  = Net::HTTP::Get.new('/')
+    post = Net::HTTP::Post.new('/')
+
+    refute @http.retry_change_requests
+
+    assert @http.can_retry?(get)
+    refute @http.can_retry?(post)
+
+    @http.retry_change_requests = true
+
+    assert @http.retry_change_requests
+
+    assert @http.can_retry?(get)
+    assert @http.can_retry?(post)
   end
 
   def test_shutdown
+    ssl_conns
     c = connection
     cs = conns
     rs = reqs
+    ts = touts
 
     orig = @http
     @http = Net::HTTP::Persistent.new 'name'
@@ -818,36 +958,102 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
 
     orig.shutdown
 
-    assert c.finished?
-    refute c2.finished?
+    @http = orig
+
+    assert c.finished?, 'last-generation connection must be finished'
+    refute c2.finished?, 'present generation connection must not be finished'
 
-    refute_same cs, conns
     refute_same rs, reqs
+    refute_same ts, touts
+
+    assert_empty conns
+    assert_empty ssl_conns
+
+    assert_empty reqs
+    assert_empty touts
   end
 
-  def test_shutdown_not_started
-    c = basic_connection
-    def c.finish() raise IOError end
+  def test_shutdown_in_all_threads
+    conns
+    ssl_conns
 
-    conns["#{@uri.host}:#{@uri.port}"] = c
+    t = Thread.new do
+      c = connection
+      ssl_conns
+      conns
+      reqs
 
-    @http.shutdown
+      Thread.stop
+
+      c
+    end
+
+    Thread.pass until t.status == 'sleep'
+
+    c = connection
+
+    assert_nil @http.shutdown_in_all_threads
+
+    assert c.finished?, 'connection in same thread must be finished'
+
+    assert_empty Thread.current[@http.generation_key]
 
-    assert_nil Thread.current[@http.connection_key]
     assert_nil Thread.current[@http.request_key]
+
+    t.run
+    assert t.value.finished?, 'connection in other thread must be finished'
+
+    assert_empty t[@http.generation_key]
+
+    assert_nil t[@http.request_key]
   end
 
   def test_shutdown_no_connections
+    conns
+    ssl_conns
+
     @http.shutdown
 
-    assert_nil Thread.current[@http.connection_key]
+    assert_empty Thread.current[@http.generation_key]
+    assert_empty Thread.current[@http.ssl_generation_key]
+
     assert_nil Thread.current[@http.request_key]
+    assert_nil Thread.current[@http.timeout_key]
+  end
+
+  def test_shutdown_not_started
+    ssl_conns
+
+    c = basic_connection
+    def c.finish() raise IOError end
+
+    conns[0]["#{@uri.host}:#{@uri.port}"] = c
+
+    @http.shutdown
+
+    assert_empty Thread.current[@http.generation_key]
+    assert_empty Thread.current[@http.ssl_generation_key]
+
+    assert_nil Thread.current[@http.request_key]
+    assert_nil Thread.current[@http.timeout_key]
+  end
+
+  def test_shutdown_ssl
+    @uri = URI 'https://example'
+
+    @http.connection_for @uri
+
+    @http.shutdown
+
+    assert_empty ssl_conns
   end
 
   def test_shutdown_thread
     t = Thread.new do
       c = connection
       conns
+      ssl_conns
+
       reqs
 
       Thread.stop
@@ -865,35 +1071,10 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
 
     t.run
     assert t.value.finished?
-    assert_nil t[@http.connection_key]
-    assert_nil t[@http.request_key]
-  end
-
-  def test_shutdown_in_all_threads
-    t = Thread.new do
-      c = connection
-      conns
-      reqs
-
-      Thread.stop
-
-      c
-    end
-
-    Thread.pass until t.status == 'sleep'
-
-    c = connection
-
-    assert_nil @http.shutdown_in_all_threads
-
-    assert c.finished?
-    assert_nil Thread.current[@http.connection_key]
-    assert_nil Thread.current[@http.request_key]
-
-    t.run
-    assert t.value.finished?
-    assert_nil t[@http.connection_key]
+    assert_empty t[@http.generation_key]
+    assert_empty t[@http.ssl_generation_key]
     assert_nil t[@http.request_key]
+    assert_nil t[@http.timeout_key]
   end
 
   def test_ssl
@@ -932,7 +1113,7 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_equal store, c.cert_store
   end
 
-  def test_default_cert_store
+  def test_ssl_cert_store_default
     @http.verify_mode = OpenSSL::SSL::VERIFY_PEER
 
     c = Net::HTTP.new 'localhost', 80
@@ -965,16 +1146,72 @@ class TestNetHttpPersistent < MiniTest::Unit::TestCase
     assert_equal OpenSSL::SSL::VERIFY_NONE, c.verify_mode
   end
 
-  def test_can_retry_change_requests
-    get  = Net::HTTP::Get.new('/')
-    post = Net::HTTP::Post.new('/')
-    assert @http.can_retry?(get)
-    refute @http.retry_change_requests
-    refute @http.can_retry?(post)
-    @http.retry_change_requests = true
-    assert @http.can_retry?(get)
-    assert @http.retry_change_requests
-    assert @http.can_retry?(post)
+  def test_ssl_warning
+    orig_verify_peer = OpenSSL::SSL::VERIFY_PEER
+    OpenSSL::SSL.send :remove_const, :VERIFY_PEER
+    OpenSSL::SSL.send :const_set, :VERIFY_PEER, OpenSSL::SSL::VERIFY_NONE
+
+    c = Net::HTTP.new 'localhost', 80
+
+    out, err = capture_io do
+      @http.ssl c
+    end
+
+    assert_empty out
+
+    assert_match %r%localhost:80%, err
+    assert_match %r%I_KNOW_THAT_OPENSSL%, err
+
+    Object.send :const_set, :I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG, nil
+
+    assert_silent do
+      @http.ssl c
+    end
+  ensure
+    OpenSSL::SSL.send :remove_const, :VERIFY_PEER
+    OpenSSL::SSL.send :const_set, :VERIFY_PEER, orig_verify_peer
+    if Object.const_defined?(:I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG) then
+      Object.send :remove_const, :I_KNOW_THAT_OPENSSL_VERIFY_PEER_EQUALS_VERIFY_NONE_IS_WRONG
+    end
+  end
+
+  def test_ssl_cleanup
+    uri1 = URI.parse 'https://one.example'
+    uri2 = URI.parse 'https://two.example'
+
+    c1 = @http.connection_for uri1
+
+    touts[c1.object_id] = Time.now
+    reqs[c1.object_id] = 5
+
+    @http.reconnect_ssl
+
+    @http.ssl_cleanup @http.ssl_generation
+
+    assert_empty ssl_conns
+    assert_empty touts
+    assert_empty reqs # sanity check, performed by #finish
+  end
+
+  def test_ssl_version_equals
+    @http.ssl_version = :ssl_version
+
+    assert_equal :ssl_version, @http.ssl_version
+    assert_equal 1, @http.ssl_generation
+  end if RUBY_VERSION > '1.9'
+
+  def test_verify_callback_equals
+    @http.verify_callback = :verify_callback
+
+    assert_equal :verify_callback, @http.verify_callback
+    assert_equal 1, @http.ssl_generation
+  end
+
+  def test_verify_mode_equals
+    @http.verify_mode = :verify_mode
+
+    assert_equal :verify_mode, @http.verify_mode
+    assert_equal 1, @http.ssl_generation
   end
 
 end
